#
# Licensed under 3-Clause BSD license available in the License file. Copyright (c) 2020-2022 iRobot Corporation. All rights reserved.
#

from .utils import bound


class Color:
    # Color sensor contants.
    ANY = -1
    WHITE = 0
    RED = 1
    GREEN = 2
    BLUE = 3
    BLACK = 4

    def __init__(self, red: int, green: int, blue: int):
        """ red, green and blue parameters must be between 0 and 255 """
        self.red = bound(red, 0, 255)
        self.green = bound(green, 0, 255)
        self.blue = bound(blue, 0, 255)



#
# Licensed under 3-Clause BSD license available in the License file. Copyright (c) 2022 iRobot Corporation. All rights reserved.
#

import asyncio
from irobot_edu_sdk.backend.backend import Backend
from irobot_edu_sdk.packet import Packet
from worker_comm import ble_write_packet, ble_disconnect, stop_program, debug_println


class Bluetooth(Backend):
    # TODO: The whole dictionary-based system, here works for devices that do not specify anything about the physical device with wich they want to connect. All this will change once the web version gets integrated into the Flutter app-
    _ble_devices = {}
    _can_write_subscribers = []

    @staticmethod
    def bluetooth_add_device(device_id):
        if not device_id in Bluetooth._ble_devices:
            Bluetooth._ble_devices[device_id] = None

    @staticmethod
    def bluetooth_data_reception(device_id, service_id, characteristic_id, data):
        # Only adds a device if it was not previously added. It's a safer (and automatic) way of dealing with devices created on the JS-side.
        if device_id in Bluetooth._ble_devices:
            device = Bluetooth._ble_devices[device_id]
            if (device != None):
                device['callback'](data)  # Data reception callback.

    # TODO: Add Multi-device support based on the ble.js functionality of registering a canWrite per device/service/characteristic key (see the Dart version).

    @staticmethod
    def bluetooth_can_write():
        for subscriber in Bluetooth._can_write_subscribers:
            can_write = getattr(subscriber, 'can_write', None)
            if callable(can_write):
                can_write()

    # TODO: Name functionality not implemented.
    # TODO: The address param from the desktop SDK may never be implemented in the web version, so document accordingly.
    def __init__(self, name: str = None):
        self.id = ''
        self.can_write_lock = asyncio.Lock()
        Bluetooth._can_write_subscribers.append(self)
        self.DEFAULT_TIMEOUT = 0.5

    async def is_connected(self) -> bool:
        # TODO.
        # print('Bluetooth.is_connected')
        return True

    async def connect(self):
        # TODO.
        # print('Bluetooth.connect')
        pass

    async def disconnect(self):
        ble_disconnect(self.id)

    def can_write(self):
        try:
            self.can_write_lock.release()
        finally:
            pass

    async def write_packet(self, packet: Packet):
        await self.can_write_lock.acquire()
        await ble_write_packet(packet.to_bytearray())
        # TODO: Evaluate if a timeout system for releasing the lock will be added.

    # Not implemented.
    # async def read_packet(self) -> Packet

    def on_data_reception(self, callback):
        for id in Bluetooth._ble_devices:
            device = Bluetooth._ble_devices[id]
            if device == None:
                Bluetooth._ble_devices[id] = {'callback': callback, 'can_write': self.can_write}  # TODO: Use constants for the keys.
                self.id = id
                break  # TODO: For now it just gets the first one that has not been assigned.

    # TODO: Evaluate if this will be here, or moved to the robot, specially for the mutlidevice system.
    def stop_program(self):
        """Extra functionality not strictly linked to Bluetooth. But, being the Bluetooth library the one that will be loaded based on the platform where the system is running (i.e., web/desktop), then this is the place to provide this function. Basically, the idea is that robots must stop the program when their stop button is pressed. On desktop native systems, that can be handled by the robot iself, but on web, the Bluetooth library is the one that communicates with the web worker.
        """
        stop_program()



#
# Licensed under 3-Clause BSD license available in the License file. Copyright (c) 2020-2022 iRobot Corporation. All rights reserved.
#

try:
    import asyncio
    from typing import Union, Dict, Tuple, Callable, Awaitable, List
except ImportError:
    import uasyncio as asyncio

from struct import pack, unpack
from .completer import Completer
from .packet import Packet
from .utils import bound, is_web
from .color import Color
from .backend.backend import Backend
from .event import Event
from .getter_types import Bumpers, TouchSensors, CliffSensor, MotorStall, Battery
import signal
import sys


def _exit_handler(signal, frame):
    print('Caught keyboard interrupt, program stopping.')
    sys.exit(0)


class Robot:
    """Base class for mobile robots."""
    DEFAULT_TIMEOUT = 3

    # Speed.
    MAX_SPEED = 500  # cm/s

    # Direction.
    DIR_LEFT = 0
    DIR_RIGHT = 1

    # LED lights.
    LIGHT_OFF = 0
    LIGHT_ON = 1
    LIGHT_BLINK = 2
    LIGHT_SPIN = 3

    # For the stall event.
    MOTOR_LEFT = 0
    MOTOR_RIGHT = 1
    MOTOR_MARKER = 2  # Reserved for robots with a physical marker.
    ERROR_NO_STALL = 0

    ERROR_OVER_CURRENT = 1
    ERROR_UNDER_CURRENT = 2
    ERROR_UNDER_SPEED = 3
    ERROR_SATURATED_PID = 4
    ERROR_TIMEOUT = 5

    def __init__(self, backend: Backend):
        self._backend = backend
        self.on_data_reception = getattr(self._backend, 'on_data_reception', None)  # Events based backend?
        if callable(self.on_data_reception):
            self.on_data_reception(self.data_reception)

        self._inc = 0
        self._run = False
        self._disable_motors = False
        self._loop = asyncio.get_event_loop()
        self._responses: Dict[Tuple[int, int, int], Completer] = {}

        self._events = {
            # (dev, cmd): event_handler(packet)
            (0, 4): self._when_stop_button_handler,
            (1, 29): self._when_motor_stalled_handler,
            (12, 0): self._when_bumped_handler,
            (14, 0): self._when_battery_handler,
            (17, 0): self._when_touched_handler,
            (20, 0): self._when_cliff_sensor_handler,
        }
        self._when_play: list[Event] = []
        self._when_stop_button: list[Event] = []
        self._when_motor_stalled: list[Event] = []
        self._when_bumped: list[Event] = []
        self._when_battery: list[Event] = []
        self._when_touched: list[Event] = []
        self._when_cliff_sensor: list[Event] = []

        # Automatically updated getters.
        self.motor_stall = MotorStall()
        self.bumpers = Bumpers()
        self.battery = Battery()
        self.touch_sensors = TouchSensors()
        self.cliff_sensor = CliffSensor()

        self.sound_enabled = True

        signal.signal(signal.SIGINT, _exit_handler)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    async def _finished(self):
        """Ensures that the robot resets its state when the program is finished neatly."""
        if self._run and await self._backend.is_connected():
            await self.stop()
        await self._backend.disconnect()

    @property
    def inc(self):
        """Access then increment wrapping id."""
        inc = self._inc
        self._inc += 1
        if self._inc > 255:
            self._inc = 0
        return inc

    def data_reception(self, data):
        # Only tries to run tasks triggered by BLE events if the program is running.
        if self._run:
            # print('🦋 ', data) # Debug.
            packet = Packet.from_bytes(bytes(data))
            self._decode_packet(packet)

    def _decode_packet(self, packet):
        """"A received packet can either be an event or a response to a command. The callback function for an event is run in a new async coroutine. A command response unblocks the command's async coroutine with received data."""
        # Check CRC.
        if not packet.check_crc():
            return

        # Check if packet is an event.
        key = (packet.dev, packet.cmd)
        if key in self._events.keys():
            self._loop.create_task(self._events[key](packet))
            return

        # Check if packet is a command response.
        key = (packet.dev, packet.cmd, packet.inc)
        if key in self._responses.keys():
            completer = self._responses.pop(key)
            completer.complete(packet)
            return

        # Unknown packet type if we got to here.

    async def _read_packets(self):
        """Reads and parses packets from robot."""
        while self._run and await self._backend.is_connected():
            await asyncio.sleep(0)  # Yield.
            packet = await self._backend.read_packet()
            self._decode_packet(packet)

    async def _main(self):
        # Connect to robot.
        if not await self._backend.is_connected():
            await self._backend.connect()
        self._run = True

        # Always resets the robot's state before starting the user's program.
        await self.stop()

        # The when_play event is always triggered first.
        for event in self._when_play:
            if not event.is_running:
                # print(event.task)
                self._loop.create_task(event.task(self))

        # Only in systems that are not events based, the packets must be polled.
        if not callable(self.on_data_reception):
            await self._read_packets()

    # Event Handlers.

    async def _when_stop_button_handler(self, packet: Packet):
        self._run = False
        stop_program = getattr(self._backend, 'stop_program', None)  # Events based backend?
        if callable(stop_program):
            stop_program()
        for event in self._when_stop_button:
            await event.run(self)

    async def _when_motor_stalled_handler(self, packet: Packet):
        self._disable_motors = True
        for event in self._when_motor_stalled:
            self.motor_stall.motor = packet.payload[4]
            self.motor_stall.cause = packet.payload[5]
            await event.run(self)

    async def _when_bumped_handler(self, packet: Packet):
        for event in self._when_bumped:
            if len(packet.payload) > 4:
                self.bumpers.left = packet.payload[4] & 0x80 != 0
                self.bumpers.right = packet.payload[4] & 0x40 != 0

                # An empty condition list means to trigger the event on every occurrence.
                if (not event.condition and self.bumpers.left) or (not event.condition and self.bumpers.right):  # Any.
                    await event.run(self)
                    continue
                if len(event.condition) > 1 and ((event.condition[0] and self.bumpers.left) or (event.condition[1] and self.bumpers.right)):
                    await event.run(self)

    async def _when_battery_handler(self, packet: Packet):
        for event in self._when_battery:
            self.battery.millivolts = unpack('>H', packet.payload[4:6])[0]
            self.battery.percent = packet.payload[6]
            # TODO: Add trigger? Probably not necessary.
            await event.run(self)

    async def _when_touched_handler(self, packet: Packet):
        for event in self._when_touched:
            if len(packet.payload) > 4:
                self.touch_sensors.front_left = packet.payload[4] & 0x80 != 0
                self.touch_sensors.front_right = packet.payload[4] & 0x40 != 0
                self.touch_sensors.back_right = packet.payload[4] & 0x20 != 0
                self.touch_sensors.back_left = packet.payload[4] & 0x10 != 0
                # An empty condition list means to trigger the event on every occurrence.
                not_condition = not event.condition
                any = (not_condition and self.touch_sensors.front_left) or (not_condition and self.touch_sensors.front_right) or (
                    not_condition and self.touch_sensors.back_left) or (not_condition and self.touch_sensors.back_right)
                if any:
                    await event.run(self)
                elif len(event.condition) > 1 and len(event.condition) < 3:
                    if any or ((event.condition[0] and self.touch_sensors.front_left) or
                               (event.condition[1] and self.touch_sensors.front_right)):
                        await event.run(self)
                elif len(event.condition) > 3:
                    if any or ((event.condition[0] and self.touch_sensors.front_left) or
                               (event.condition[1] and self.touch_sensors.front_right) or
                               (event.condition[2] and self.touch_sensors.back_left) or
                               (event.condition[3] and self.touch_sensors.back_right)):
                        await event.run(self)

    async def _when_cliff_sensor_handler(self, packet: Packet):
        self.cliff_sensor.disable_motors = packet.payload[4] != 0
        for event in self._when_cliff_sensor:
            # TODO: Add trigger
            await event.run(self)

    # Event Callbacks.

    def when_play(self, callback: Callable[[], Awaitable[None]]):
        """Register when play callback of type: async def fn()."""
        self._when_play.append(Event(True, callback))

    def when_stop(self, callback: Callable[[], Awaitable[None]]):
        """Register when stop callback of type async def fn()."""
        self._when_stop_button.append(Event(True, callback))

    def when_motor_stalled(self, condition: list[int, int], callback: Callable[[MotorStall], Awaitable[None]]):
        """Register when motor stall callback of type async def fn(motor: Motor, stall: Stall)."""
        self._when_motor_stalled.append(Event(condition, callback))

    def when_bumped(self, condition: list[bool, bool], callback: Callable[[Bumpers], Awaitable[None]]):
        """Register when bumper callback of type: async def fn(left: bool, right: bool)."""
        self._when_bumped.append(Event(condition, callback))

    def when_battery(self, condition: list[int, int], callback: Callable[[Battery], Awaitable[None]]):
        """Register when battery callback of type: async def fn(mV: int, percent: int)."""
        self._when_battery.append(Event(condition, callback))

    def when_touched(self, condition: list[bool, bool, bool, bool], callback: Callable[[TouchSensors], Awaitable[None]]):
        """Register when touch callback of type: async def fn(front_left: bool, front_right: bool, back_left: bool, back_right: bool)."""
        self._when_touched.append(Event(condition, callback))

    def when_cliff_sensor(self, condition: list[bool], callback: Callable[[bool], Awaitable[None]]):
        """Register when cliff callback of type: async def fn(over_cliff: bool)."""
        self._when_cliff_sensor.append(Event(condition, callback))

    # Commands.

    def play(self):
        """Start the program."""
        if self._run:
            # Calling play() more than once makes the program unpredicable.
            print('🟧 Robot program already running')
            return

        try:
            if hasattr(self._loop, 'is_running') and self._loop.is_running():
                self._main_task = self._loop.create_task(self._main())
            else:
                self._loop.run_until_complete(self._main())
        except KeyboardInterrupt:
            print('Caught keyboard interrupt exception, program stopping.')
            self._run = False
        except SystemExit:
            self._run = False
        finally:
            # This fails on the web version, so determining the platform is crucial:
            if not is_web():
                for task in asyncio.all_tasks(self._loop):
                    task.cancel()

            if not hasattr(self._loop, 'is_running') or not self._loop.is_running():
                self._loop.run_until_complete(self._finished())
                self._loop.close()

    async def stop(self):
        """Stop and reset robot."""
        await self._backend.write_packet(Packet(0, 3, self.inc))

    # TODO: Evaluate if this one needs to be async (most likely not)
    def stop_all_events(self):
        # TODO
        pass

    # TODO: Evaluate if this one needs to be async (most likely not)
    def stop_other_events(self):
        # TODO
        pass

    async def stop_sound(self):
        """Stop currently playing note."""
        await self._backend.write_packet(Packet(5, 1, self.inc))
        self.sound_enabled = False

    async def wait(self, time: Union[int, float]):
        await asyncio.sleep(time)

    async def get_versions(self, board: int) -> List[int]:
        """Get version numbers. Returns [board, fw maj, fw min, hw maj, hw min, boot maj, boot min, proto maj, proto min]."""
        dev, cmd, inc = 0, 0, self.inc
        completer = Completer()
        self._responses[(dev, cmd, inc)] = completer
        await self._backend.write_packet(Packet(dev, cmd, inc, bytes([board])))
        packet = await completer.wait(self.DEFAULT_TIMEOUT)
        return packet.payload[: 9] if packet else []

    async def set_name(self, name: str):
        """Set robot name."""
        utf = name.encode('utf-8')
        while len(utf) > Packet.PAYLOAD_LEN:
            name = name[: -1]
            utf = name.encode('utf-8')
        await self._backend.write_packet(Packet(0, 1, self.inc, utf))

    async def get_name(self) -> str:
        """Get robot name."""
        dev, cmd, inc = 0, 2, self.inc
        completer = Completer()
        self._responses[(dev, cmd, inc)] = completer
        await self._backend.write_packet(Packet(dev, cmd, inc))
        packet = await completer.wait(self.DEFAULT_TIMEOUT)
        return packet.payload.decode('utf-8').rstrip('\0') if packet else ''

    async def disconnect(self):
        """Disconnect Bluetooth from robot side."""
        await self._backend.write_packet(Packet(0, 6, self.inc))

    async def enable_events(self, bitfield: bytes):
        """Enable notifications for events. Accepts 128-bit bitfield for devices 0 to 127."""
        await self._backend.write_packet(Packet(0, 7, self.inc, bitfield))

    async def disable_events(self, bitfield: bytes):
        """Disable notifications for events. Accepts 128-bit bitfield for devices 0 to 127."""
        await self._backend.write_packet(Packet(0, 9, self.inc, bitfield))

    async def get_enabled_events(self) -> bytes:
        """Return 128-bit bitfield for devices 0 to 127."""
        dev, cmd, inc = 0, 11, self.inc
        completer = Completer()
        self._responses[(dev, cmd, inc)] = completer
        await self._backend.write_packet(Packet(dev, cmd, inc))
        packet = await completer.wait(self.DEFAULT_TIMEOUT)
        return packet.payload if packet else bytes()

    async def get_serial_number(self) -> str:
        """Get serial number string."""
        dev, cmd, inc = 0, 14, self.inc
        completer = Completer()
        self._responses[(dev, cmd, inc)] = completer
        await self._backend.write_packet(Packet(dev, cmd, inc))
        packet = await completer.wait(self.DEFAULT_TIMEOUT)
        try:
            return packet.payload.decode('utf-8').rstrip('\0') if packet else ''
        except UnicodeDecodeError:
            return ''.join([format(b, "02X") for b in packet.payload])

    async def get_sku(self) -> str:
        """Get robot type SKU string."""
        dev, cmd, inc = 0, 15, self.inc
        completer = Completer()
        self._responses[(dev, cmd, inc)] = completer
        await self._backend.write_packet(Packet(dev, cmd, inc))
        packet = await completer.wait(self.DEFAULT_TIMEOUT)
        return packet.payload.decode('utf-8').rstrip('\0') if packet else ''

    async def get_battery_level(self) -> Tuple[int, int]:
        # TODO: Update robot's getter.
        # Get battery level. Returns (mV, percent)
        dev, cmd, inc = 14, 1, self.inc
        completer = Completer()
        self._responses[(dev, cmd, inc)] = completer
        await self._backend.write_packet(Packet(dev, cmd, inc))
        packet = await completer.wait(self.DEFAULT_TIMEOUT)
        return (unpack(">H", packet.payload[4:6])[0], packet.payload[6]) if packet else (0, 0)

    async def set_wheel_speeds(self, left: Union[int, float], right: Union[int, float]):
        """Set motor speed in cm/s."""
        if self._disable_motors:
            return
        left = bound(int(left * 10), -self.MAX_SPEED, self.MAX_SPEED)
        right = bound(int(right * 10), -self.MAX_SPEED, self.MAX_SPEED)
        await self._backend.write_packet(Packet(1, 4, self.inc, pack('>ii', left, right)))

    async def set_left_speed(self, speed: Union[int, float]):
        """Set left motor speed in cm/s."""
        if self._disable_motors:
            return
        speed = bound(int(speed * 10), -self.MAX_SPEED, self.MAX_SPEED)
        await self._backend.write_packet(Packet(1, 6, self.inc, pack('>i', speed)))

    async def set_right_speed(self, speed: Union[int, float]):
        """Set right motor speed in cm/s."""
        if self._disable_motors:
            return
        speed = bound(int(speed * 10), -self.MAX_SPEED, self.MAX_SPEED)
        await self._backend.write_packet(Packet(1, 7, self.inc, pack('>i', speed)))

    async def move(self, distance: Union[int, float]):
        """Drive distance in centimeters."""
        if self._disable_motors:
            return
        dev, cmd, inc = 1, 8, self.inc
        packet = Packet(dev, cmd, inc, pack('>i', int(distance * 10)))
        completer = Completer()
        self._responses[(dev, cmd, inc)] = completer
        await self._backend.write_packet(packet)
        await completer.wait(self.DEFAULT_TIMEOUT + int(abs(distance) / 10))

    async def turn_right(self, angle: Union[int, float]):
        """Rotate angle in degrees."""
        if self._disable_motors:
            return
        dev, cmd, inc = 1, 12, self.inc
        packet = Packet(dev, cmd, inc, pack('>i', int(angle * 10)))
        completer = Completer()
        self._responses[(dev, cmd, inc)] = completer
        await self._backend.write_packet(packet)
        await completer.wait(self.DEFAULT_TIMEOUT + int(abs(angle) / 100))

    async def turn_left(self, angle: Union[int, float]):
        await self.turn_right(-angle)

    async def arc(self, direction: int, angle: Union[int, float], radius: Union[int, float]):
        """Drive arc defined by angle in degrees and radius in cm."""
        if self._disable_motors:
            return
        dev, cmd, inc = 1, 27, self.inc
        if direction == Robot.DIR_LEFT:
            angle = -angle
            radius = -radius
        payload = pack('>ii', int(angle * 10), int(radius * 10))
        completer = Completer()
        self._responses[(dev, cmd, inc)] = completer
        await self._backend.write_packet(Packet(dev, cmd, inc, payload))
        await completer.wait(self.DEFAULT_TIMEOUT + int(abs(radius * angle / 573)))

    async def set_lights(self, animation: int, color: Color):
        """Set roobt's LEDs to animation with color red, green, blue."""
        animation = bound(animation, Robot.LIGHT_OFF, Robot.LIGHT_SPIN)
        color.red = bound(color.red, 0, 255)
        color.green = bound(color.green, 0, 255)
        color.blue = bound(color.blue, 0, 255)
        payload = bytes([animation, color.red, color.green, color.blue])
        await self._backend.write_packet(Packet(3, 2, self.inc, payload))

    async def set_lights_rgb(self, red: int, green: int, blue: int):
        await self.set_lights(Robot.LIGHT_ON, Color(red, green, blue))

    async def play_note(self, frequency: Union[float, int], duration: Union[float, int]):
        """Play note with frequency in hertz for duration in seconds."""
        dev, cmd, inc = 5, 0, self.inc
        payload = pack('>IH', abs(frequency), abs(int(duration * 1000)))
        completer = Completer()
        self._responses[(dev, cmd, inc)] = completer
        await self._backend.write_packet(Packet(dev, cmd, inc, payload))
        await completer.wait(self.DEFAULT_TIMEOUT + int(abs(duration)))

    # async def play_sweep(
    #    self,
    #    initialFreq: Union[float, int],
    #    finalFreq: Union[float, int],
    #    duration: Union[float, int],
    #    attack: int = 10,
    #    release: int = 10,
    #    volume: int = 255,
    #    modulation_type: int = 2,
    #    modulation_rate: int = 200,
    #    append: int = 0, # TODO: Must be bool.
    #    ):
    #    """Play sweep from initialFreq to finalFreq (both in hertz), for duration (in seconds)."""
    #    dev, cmd, inc = 5, 5, self.inc
    #    payload = pack(
    #        '>IIHBBBBBB',
    #        abs(initialFreq),
    #        abs(finalFreq),
    #        abs(int(duration * 1000)),
    #        attack,
    #        release,
    #        volume,
    #        modulation_type,
    #        modulation_rate,
    #        append
    #    )
    #    completer = Completer()
    #    self._responses[(dev, cmd, inc)] = completer
    #    await self._backend.write_packet(Packet(dev, cmd, inc, payload))
    #    await completer.wait(self.DEFAULT_TIMEOUT + int(duration))

    async def say(self, phrase: str):
        """Say a phrase in robot language."""
        self.sound_enabled = True
        buf = phrase.encode('utf-8')
        for payload in [
                buf[i:i + Packet.PAYLOAD_LEN]
                for i in range(0, len(buf), Packet.PAYLOAD_LEN)
        ]:
            if self.sound_enabled:
                dev, cmd, inc = 5, 4, self.inc
                completer = Completer()
                self._responses[(dev, cmd, inc)] = completer
                await self._backend.write_packet(Packet(dev, cmd, inc, payload))
                await completer.wait(self.DEFAULT_TIMEOUT + len(payload))
                break



#
# Licensed under 3-Clause BSD license available in the License file. Copyright (c) 2020-2022 iRobot Corporation. All rights reserved.
#

import math
from typing import Union
from struct import pack, unpack
from .backend.backend import Backend
from .completer import Completer
from .packet import Packet
from .utils import bound
from .getter_types import IPv4Addresses, IrProximity, Pose
from irobot_edu_sdk.robot import Robot


class Create3(Robot):
    """Create 3 robot object."""

    DOCK_STATUS_SUCCEEDED = 0
    DOCK_STATUS_ABORTED   = 1
    DOCK_STATUS_CANCELED  = 2
    DOCK_RESULT_UNDOCKED = 0
    DOCK_RESULT_DOCKED   = 1

    def __init__(self, backend: Backend):
        super().__init__(backend=backend)

        # Getters.
        self.ipv4_address = IPv4Addresses()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    async def get_ipv4_address(self) -> IPv4Addresses:
        """Get the robot's ipv4 address as a IPv4Addresses, which contains wlan0, wlan1 and usb0. Returns None if anything went wrong."""
        dev, cmd, inc = 100, 1, self.inc
        completer = Completer()
        self._responses[(dev, cmd, inc)] = completer
        await self._backend.write_packet(Packet(dev, cmd, inc))
        packet = await completer.wait(self.DEFAULT_TIMEOUT)
        if packet:
            self.ipv4_address.wlan0 = [packet.payload[0], packet.payload[1], packet.payload[2], packet.payload[3]]
            self.ipv4_address.wlan1 = [packet.payload[4], packet.payload[5], packet.payload[6], packet.payload[7]]
            self.ipv4_address.usb0 = [packet.payload[8], packet.payload[9], packet.payload[10], packet.payload[11]]
            return self.ipv4_address
        return None

    async def get_ir_proximity(self):
        dev, cmd, inc = 11, 1, self.inc
        completer = Completer()
        self._responses[(dev, cmd, inc)] = completer
        await self._backend.write_packet(Packet(dev, cmd, inc))
        packet = await completer.wait(self.DEFAULT_TIMEOUT)
        if packet:
            unpacked = unpack('>IHHHHHH', packet.payload)
            ir_proximity = IrProximity()
            ir_proximity.sensors = list(unpacked[1:])
            return ir_proximity
        return None

    async def get_packed_ir_proximity(self):
        """Get Packed IR Proximity Values and States"""
        dev, cmd, inc = 11, 2, self.inc
        completer = Completer()
        self._responses[(dev, cmd, inc)] = completer
        await self._backend.write_packet(Packet(dev, cmd, inc))
        packet = await completer.wait(self.DEFAULT_TIMEOUT)
        if packet:
            payload = packet.payload
            timestamp = unpack('>I', payload[0:4])[0]
            ir_proximity = IrProximity()
            #ir_proximity.state = payload[4]
            ir_proximity.sensors = [
                (payload[ 5] << 4) + (payload[12] >> 4),
                (payload[ 6] << 4) + (payload[12] & 0xF),
                (payload[ 7] << 4) + (payload[13] >> 4),
                (payload[ 8] << 4) + (payload[13] & 0xF),
                (payload[ 9] << 4) + (payload[14] >> 4),
                (payload[10] << 4) + (payload[14] & 0xF),
                (payload[11] << 4) + (payload[15] >> 4),
            ]
            return ir_proximity
        return None

    async def get_position(self):
        dev, cmd, inc = 1, 16, self.inc
        completer = Completer()
        self._responses[(dev, cmd, inc)] = completer
        await self._backend.write_packet(Packet(dev, cmd, inc))
        packet = await completer.wait(self.DEFAULT_TIMEOUT)
        if packet:
            payload = packet.payload
            timestamp = unpack('>I', payload[0:4])[0]
            x = unpack('>i', payload[4:8])[0]
            y = unpack('>i', payload[8:12])[0]
            heading = unpack('>h', payload[12:14])[0] / 10
            return Pose(x, y, heading)
        return None

    async def reset_navigation(self):
        await self._backend.write_packet(Packet(1, 15, self.inc))

    async def navigate_to(self, x: Union[int, float], y: Union[int, float], heading: Union[int, float] = None):
        """ If heading is None, then it will be ignored, and the robot will arrive to its destination
        pointing towards the direction of the line between the destination and the origin points.
        Units:
            x, y: cm
            heading: deg
        """
        if self._disable_motors:
            return
        dev, cmd, inc = 1, 17, self.inc
        _heading = -1
        if heading is not None:
            _heading = int(heading * 10)
            _heading = bound(_heading, 0, 3599)
        payload = pack('>iih', int(x * 10), int(y * 10), _heading)
        completer = Completer()
        self._responses[(dev, cmd, inc)] = completer
        await self._backend.write_packet(Packet(dev, cmd, inc, payload))
        timeout = self.DEFAULT_TIMEOUT + int(math.sqrt(x * x + y * y) / 10) + 4  # 4 is the timeout for a potential rotation.
        await completer.wait(timeout)

    async def dock(self):
        """Request a docking action."""
        dev, cmd, inc = 1, 19, self.inc
        completer = Completer()
        self._responses[(dev, cmd, inc)] = completer
        await self._backend.write_packet(Packet(dev, cmd, inc))
        packet = await completer.wait(60)
        if packet:
            unpacked = unpack('>IBBHHHHH', packet.payload)
            return {'timestamp': unpacked[0], 'status': unpacked[1], 'result': unpacked[2]}
        return None

    async def undock(self):
        """Request an undocking action."""
        dev, cmd, inc = 1, 20, self.inc
        completer = Completer()
        self._responses[(dev, cmd, inc)] = completer
        await self._backend.write_packet(Packet(dev, cmd, inc))
        packet = await completer.wait(30)
        if packet:
            unpacked = unpack('>IBBHHHHH', packet.payload)
            return {'timestamp': unpacked[0], 'status': unpacked[1], 'result': unpacked[2]}
        return None

    async def get_docking_values(self):
        """Get docking values."""
        dev, cmd, inc = 19, 1, self.inc
        completer = Completer()
        self._responses[(dev, cmd, inc)] = completer
        await self._backend.write_packet(Packet(dev, cmd, inc))
        packet = await completer.wait(self.DEFAULT_TIMEOUT)
        if packet:
            unpacked = unpack('>IBBBBHHHH', packet.payload)
            return {'timestamp': unpacked[0], 'contacts': unpacked[1], 'IR sensor 0': unpacked[2],
                    'IR sensor 1': unpacked[3], 'IR sensor 2': unpacked[4]}
        return None



#
# Licensed under 3-Clause BSD license available in the License file. Copyright (c) 2020-2022 iRobot Corporation. All rights reserved.
#

from typing import Union
import sys

def bound(value: Union[int, float], low: Union[int, float], high: Union[int, float]) -> int:
    _value, _low, _high = int(value), int(low), int(high)
    return min(_high, max(_low, _value))

def stop_program():
    # Need to do something to catch the SystemExit exception.
    sys.exit()

# Returns True if running on a web (Pyodide) environment.
def is_web():
    return sys.platform == 'emscripten'



#
# Licensed under 3-Clause BSD license available in the License file. Copyright (c) 2020-2022 iRobot Corporation. All rights reserved.
#

try:
    from asyncio import sleep
    from time import time
    from typing import Any, Optional
except ImportError:
    from uasyncio import sleep
    from utime import time


class Completer():
    def __init__(self):
        self.clear()

    def clear(self):
        self._flag = False
        self._data = None

    async def wait(self, timeout: Optional[int] = None) -> Optional[Any]:
        start = int(time())
        while not self._flag:
            if timeout and int(time()) - start > timeout:
                break
            await sleep(0)
        return self.value()

    def is_complete(self) -> bool:
        return self._flag

    def complete(self, data=None):
        self._flag = True
        self._data = data

    def value(self) -> Optional[Any]:
        return self._data



#
# Licensed under 3-Clause BSD license available in the License file. Copyright (c) 2020-2022 iRobot Corporation. All rights reserved.
#


import functools

try:
    import asyncio
except ImportError:
    import uasyncio as asyncio


async def hand_over():
    # Python idiomatic way of yielding control from a coroutine.
    # See https://github.com/python/asyncio/issues/284
    await asyncio.sleep(0)


class Event:
    """ This class will be used for robots other than Root, thus the naming """

    def __init__(self, condition, task):
        self.condition = condition
        self.task = task
        self.is_running = False
        # self.prev_data = None  # can be used for filtering while triggering if needed

    async def run(self, device):
        if not self.is_running:
            self.is_running = True
            await self.task(device)
            self.is_running = False


def event(method, condition=None):
    """ Convienience decorator that allows to define any function as a device event with the syntax:
        `@event(device.method, condition)`
    """
    def decorator_event(func):
        if condition != None:
            method(condition, func)
        else:
            method(func)

        @functools.wraps(func)  # May not work on MicroPython (in that case, load functools conditionally).
        def wrapper_event(*args, **kwargs):
            func(*args, **kwargs)
        return wrapper_event
    return decorator_event



#
# Licensed under 3-Clause BSD license available in the License file. Copyright (c) 2020-2022 iRobot Corporation. All rights reserved.
#

import math
from typing import Union, Callable, Awaitable, List
from struct import pack, unpack
from .backend.backend import Backend
from .event import Event
from .completer import Completer
from .packet import Packet
from .utils import bound
from .robot import Robot
from .getter_types import Pose, Movement, ColorSensor, LightSensors


class Root(Robot):
    """Create 3 robot object"""

    # Marker/eraser.
    MARKER_UP = 0
    MARKER_DOWN = 1
    MARKER_ERASE = 2

    # Light sensor.
    LIGHT_UNKNOWN = 0
    LIGHT_DARK = 4
    LIGHT_RIGHT_BRIGHTER_THAN_LEFT = 5
    LIGHT_LEFT_BRIGHTER_THAN_RIGHT = 6
    LIGHT_BRIGHT = 7

    # Gravity compensation.
    GRAVITY_OFF = 0
    GRAVITY_ON = 1
    GRAVITY_WHEN_MARKER = 2

    # Color sensor.
    COLOR_LIGHTING_OFF = 0
    COLOR_LIGHTING_RED = 1
    COLOR_LIGHTING_GREEN = 2
    COLOR_LIGHTING_BLUE = 3
    COLOR_LIGHTING_ALL = 4

    COLOR_FORMAT_ADC_COUNTS = 0
    COLOR_FORMAT_MILLIVOLTS = 1

    COLOR_SENSORS_0_TO_7 = 0
    COLOR_SENSORS_8_TO_15 = 1
    COLOR_SENSORS_16_TO_23 = 2
    COLOR_SENSORS_24_TO_31 = 3

    def __init__(self, backend: Backend):
        super().__init__(backend=backend)

        self._events[(4, 2)] = self._when_color_scanned_handler
        self._events[(13, 0)] = self._when_light_seen_handler

        self._when_color_scanned: list[Event] = []
        self._when_light_seen: list[Event] = []

        # Getters.
        self.color_sensor = ColorSensor()
        self.light_sensors = LightSensors()

        self.pose = Pose()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    # Event Handlers.

    async def _when_color_scanned_handler(self, packet: Packet):
        for event in self._when_color_scanned:
            self.color_sensor.colors = [c >> i & 0xF for c in packet.payload for i in range(4, -1, -4)]
            # TODO: Add trigger.
            await event.run(self)

    # TODO: Test.
    async def _when_light_seen_handler(self, packet: Packet):
        for event in self._when_light_seen:
            self.light_sensors.state = packet.payload[4]
            self.light_sensors.left = unpack(">H", packet.payload[5:7])[0]
            self.light_sensors.right = unpack(">H", packet.payload[7:9])[0]
            # TODO: Add trigger
            await event.run(self)

    # Event Callbacks.

    def when_color_scanned(self, condition: list[List[int]], callback: Callable[[ColorSensor], Awaitable[None]]):
        """Register when color callback of type async def fn(colors:
        List[Color])"""
        self._when_color_scanned.append(Event(condition, callback))

    def when_light_seen(self, condition: list[int, int, int], callback: Callable[[LightSensors], Awaitable[None]]):
        """Register when light callback of type: async def fn(state: Light, left_mV: int, right_mV: int)"""
        self._when_light_seen.append(Event(condition, callback))

    # Commands.

    async def stop(self):
        await self.reset_navigation()
        await super().stop()

    async def move(self, distance: Union[int, float]):
        await super().move(distance)
        self.pose.move(distance)

    async def turn_right(self, angle: Union[int, float]):
        await super().turn_right(angle)
        self.pose.turn_left(-angle)

    async def set_marker(self, position: int):
        """Set marker to position of type Marker"""
        if self._disable_motors:
            return
        dev, cmd, inc = 2, 0, self.inc
        payload = bytes([bound(position, Root.MARKER_UP, Root.MARKER_ERASE)])
        completer = Completer()
        self._responses[(dev, cmd, inc)] = completer
        await self._backend.write_packet(Packet(dev, cmd, inc, payload))
        await completer.wait(self.DEFAULT_TIMEOUT)

    async def set_gravity_compensation(self, gravity: int, amount: Union[int, float]):
        """Set vertical driving compensation for gravity and amount between 0% and 300%"""
        gravity = bound(gravity, Root.GRAVITY_OFF, Root.GRAVITY_WHEN_MARKER)
        amount = bound(int(amount * 10), 0, 3000)
        await self._backend.write_packet(Packet(1, 13, self.inc, pack(">BH", gravity, amount)))

    # It is async since this command will be implemented in the robot's firmware.
    async def get_position(self):
        return self.pose

    async def reset_navigation(self):
        self.pose.x = 0
        self.pose.y = 0
        self.pose.heading = 90

    def compute_movement_to(self, x, y):
        dx = x - self.pose.x
        dy = y - self.pose.y
        return Movement(math.sqrt(dx * dx + dy * dy), math.degrees(math.atan2(dy, dx)) - self.pose.heading)

    # TODO: Finish implementation.
    async def navigate_to(self, x: Union[int, float], y: Union[int, float], heading: Union[int, float] = None):
        """ If heading is None, then it will be ignored, and the robot will arrive to its destination
        pointing towards the direction of the line between the destination and the origin points."""
        movement = self.compute_movement_to(x, y)
        await self.turn_left(Movement.minimize_angle(movement.angle))
        await self.move(movement.distance)
        if heading is not None:
            # TODO: test.
            await self.turn_left(Movement.minimize_angle(heading - self.heading))



#
# Licensed under 3-Clause BSD license available in the License file. Copyright (c) 2020-2022 iRobot Corporation. All rights reserved.
#

# Enables single-file importing of most of the commonly used SDK objects and functions.

from .color import Color
from .event import event, hand_over
from .robot import Robot
from .root import Root
from .create3 import Create3



#
# Licensed under 3-Clause BSD license available in the License file. Copyright (c) 2020-2022 iRobot Corporation. All rights reserved.
#

import math
from typing import List

# Getter classes used both for Root's getters and for the params on
# its event callbacks. These classes are Root specific.


class Pose:
    def __init__(self, x=0, y=0, heading=90):
        self.x = x
        self.y = y
        self.heading = heading  # [deg]

    def move(self, distance):
        self.x += distance * math.cos(math.radians(self.heading))
        self.y += distance * math.sin(math.radians(self.heading))

    def turn_left(self, angle):
        self.heading += angle


class Movement:
    def __init__(self, distance=0, angle=0):
        self.distance = distance
        self.angle = self.minimize_angle(angle)  # [deg]

    @staticmethod
    def minimize_angle(angle):
        result = angle
        while result > 180:
            result -= 360
        while result < -180:
            result += 360
        return result


class IrProximity:
    def __init__(self):
        self.sensors: List[int] = []


class ColorSensor:
    def __init__(self):
        # contains the 32 areas from the real sensor
        self.colors: List[int] = []


class Bumpers:
    def __init__(self):
        self.left = False
        self.right = False


class TouchSensors:
    def __init__(self):
        self.front_left = False
        self.front_right = False
        self.back_right = False
        self.back_left = False


class CliffSensor:
    def __init__(self):
        self.disable_motors = False


class LightSensors:
    def __init__(self):
        self.state: int = 0
        self.left: int = 0
        self.right: int = 0


class IPv4Addresses:
    def __init__(self):
        self.wlan0: List[int] = []
        self.wlan1: List[int] = []
        self.usb0: List[int] = []


class MotorStall:
    def __init__(self):
        self.motor: int = 0
        self.cause: int = 0


class Battery:
    def __init__(self):
        self.millivolts: int = 0
        self.percent: int = 0



#
# Licensed under 3-Clause BSD license available in the License file. Copyright (c) 2020-2022 iRobot Corporation. All rights reserved.
#

try:
    from typing import Optional
except ImportError:
    pass
from struct import pack


class Packet():
    """robot packet type"""
    PACKET_LEN = 20
    PAYLOAD_LEN = 16

    def __init__(self,
                 dev: int,
                 cmd: int,
                 inc: int,
                 payload: bytes = bytes(PAYLOAD_LEN),
                 crc: Optional[int] = None):
        self.dev = dev
        self.cmd = cmd
        self.inc = inc
        assert len(payload) <= self.PAYLOAD_LEN, "invalid payload length"
        self.payload = payload + bytes(self.PAYLOAD_LEN - len(payload))
        self._crc = crc

    @classmethod
    def from_bytes(cls, raw: bytes):
        """create a new packet instance from raw bytes"""
        assert len(raw) == cls.PACKET_LEN, "invalid packet length"
        return Packet(raw[0], raw[1], raw[2], payload=raw[3:19], crc=raw[19])

    def to_bytes(self):
        """20 byte packet with crc"""
        return self.packet() + bytes([self.calc_crc()])

    def to_bytearray(self):
        """mutable 20 byte array with crc"""
        return bytearray(self.to_bytes())

    def packet(self):
        """19 byte packet without crc"""
        return pack("3B", self.dev, self.cmd, self.inc) + self.payload

    def check_crc(self):
        """check if computed crc matches stored crc"""
        return False if self._crc is None else self._crc == self.calc_crc()

    def calc_crc(self) -> int:
        """calculates crc for packet"""
        crc = 0x00
        for c in self.packet():
            for i in range(8):
                b = crc & 0x80
                if c & (0x80 >> i):
                    b ^= 0x80
                crc <<= 1
                if b:
                    crc ^= 0x07
            crc &= 0xFF
        return crc

    @property
    def crc(self) -> int:
        """return stored crc or calculate new crc"""
        return self._crc if self._crc is not None else self.calc_crc()



#
# Licensed under 3-Clause BSD license available in the License file. Copyright (c) 2020 iRobot Corporation. All rights reserved.
#

"""
This is a USB Serial class that implements the Backend interface methods.

It is compatible with CPython using pySerial and the MicroPython Unix port using micropython-serial.
"""

try:
    from asyncio import sleep
except ImportError:
    from uasyncio import sleep

from binascii import hexlify, unhexlify
from serial import Serial as _Serial
from .backend import Backend
from ..packet import Packet


class Serial(Backend):
    def __init__(self, port: str):
        self._serial = _Serial(port, 115200)

    async def connect(self):
        self._serial.open()

    async def is_connected(self) -> bool:
        try:
            return self._serial.isOpen()
        except AttributeError:
            return True  # no isOpen() for micropython

    async def disconnect(self):
        self._serial.close()

    async def read_packet(self) -> Packet:
        string = b''
        while not (string.endswith(b'\n') and len(string) > 40):
            await sleep(0)
            while not self._serial.inWaiting():
                await sleep(0)
            string += self._serial.read(1)
        return Packet.from_bytes(unhexlify(string[-41:-1]))

    async def write_packet(self, packet: Packet):
        string = hexlify(packet.to_bytes()) + b'\n'
        self._serial.write(string)



#
# Licensed under 3-Clause BSD license available in the License file. Copyright (c) 2022 iRobot Corporation. All rights reserved.
#

import sys
if 'pyodide' in sys.modules:
    from irobot_edu_sdk.backend.bluetooth_web import Bluetooth
else:
    from irobot_edu_sdk.backend.bluetooth_desktop import Bluetooth



#
# Licensed under 3-Clause BSD license available in the License file. Copyright (c) 2020-2022 iRobot Corporation. All rights reserved.
#

"""
This is a Bluetooth Low Energy class that implements the Backend interface methods.

It is compatible with CPython on macOS, Windows, and Linux using the Bleak library.
"""

from asyncio import sleep
from queue import SimpleQueue
from typing import Optional
from bleak import BleakClient, BleakScanner
from .backend import Backend
from ..packet import Packet


class Bluetooth(Backend):
    ROOT_ID_SERVICE = "48c5d828-ac2a-442d-97a3-0c9822b04979"
    UART_SERVICE = "6e400001-b5a3-f393-e0a9-e50e24dcca9e"
    TX_CHARACTERISTIC = "6e400002-b5a3-f393-e0a9-e50e24dcca9e"
    RX_CHARACTERISTIC = "6e400003-b5a3-f393-e0a9-e50e24dcca9e"
    UUID_KEY = 'uuids'

    def __init__(self, name: str = None, address: Optional[str] = None):
        """If no name is provided, connects to the first device found."""
        self._name = name
        self._address = address
        self._device = None
        self._client: Optional[BleakClient] = None
        self._queue: SimpleQueue = SimpleQueue()

    def rx_handler(self, characteristic, data):
        self._queue.put(Packet.from_bytes(bytes(data)))

    async def connect(self):
        """This method does not exit until a robot is found"""

        while not self._address:
            # Filter by service uuid is not supported on all operating systems.
            devices = await BleakScanner.discover(service_uuids=[self.ROOT_ID_SERVICE, self.UART_SERVICE])
            for device in devices:
                # print('device.metadata:', device.metadata) # Debug.
                if self.UUID_KEY in device.metadata:
                    if self.ROOT_ID_SERVICE in device.metadata[self.UUID_KEY]:
                        if self._name is None or self._name == device.name:
                            self._address = device.address
                            self._device = device
                            break
        print(f'Connecting to {device.name} ({device.address})')
        if self._device:
            self._client = BleakClient(self._device)
        else:
            self._client = BleakClient(self._address)

        if await self._client.connect():
            await self._client.start_notify(self.RX_CHARACTERISTIC, self.rx_handler)

    async def is_connected(self) -> bool:
        return self._client.is_connected if self._client else False

    async def disconnect(self):
        if self.is_connected:
            await self._client.disconnect()
        self._client = None

    async def read_packet(self) -> Packet:
        while self._queue.empty():
            await sleep(0)
        return self._queue.get()

    async def write_packet(self, packet: Packet):
        if self._client:
            await self._client.write_gatt_char(self.TX_CHARACTERISTIC, packet.to_bytearray(), True)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['irobot_edu_sdk', 'irobot_edu_sdk.backend']

package_data = \
{'': ['*']}

install_requires = \
['bleak>=0.19.0,<0.20.0', 'pyserial>=3.4,<4.0']

setup_kwargs = {
    'name': 'irobot-edu-sdk',
    'version': '0.1.0',
    'description': 'Python SDK for iRobot Edu robots',
    'long_description': "# iRobot® Education Python® SDK\n\nThis is an installable Python SDK for the iRobot Education robots.\nIt supports the iRobot® Root® rt0, Root® rt1 and Create® 3 educational robots.\n\nThis SDK uses a similar format as Learning Level 3 of the iRobot® Coding App ([code.irobot.com](https://code.irobot.com)) and the iRobot® Python Web Playground ([python.irobot.com](https://python.irobot.com)).\nAs such, it uses async execution and is compatible with Python® 3.9 or greater. MicroPython may work, but it is not tested.\n\nThis SDK is designed to communicate with robots via Bluetooth® Low Energy on Windows®, macOS®, and Linux® systems.\nExperimental backends are provided for communicating over USB serial and USB on MicroPython boards with native USB ports.\nThese are not currently supported.\n\n## Disclaimer: this is a BETA release.\nThe BETA version of the iRobot® Python SDK is provided “as is” to conduct testing and gather user feedback. There is no guarantee on its performance or compatibility, continued availability, or on a timeline for improvements.\n\nPlease do file [issues](https://github.com/iRobotEducation/irobot-edu-python-sdk/issues) as they are found, so that the team (or the community) can address them.\n\n\n## Installation\n\n```\ngit clone https://github.com/iRobotEducation/irobot-edu-python-sdk.git\n```\n\nFrom the newly cloned SDK's directory, run:\n\n```\npip3 install .\n```\n\n\n## Usage\n\nA robot object is created from one of the communication backends.\nThis robot object has a number of event callbacks that can be registered and commands that can be sent to the robot.\nAfter a program is setup, the `play()` method will start execution and block until the robot's nose or power button is pressed, the robot is disconnected, or `Ctrl-C` is pressed.\n\n\n## Examples\n\nA good way of getting started is by looking at the examples provided [here](https://github.com/iRobotEducation/irobot-edu-python-sdk/tree/main/examples).\n\nThese examples mirror the iRobot Education [Python Web Playground](http://python.irobot.com).\n\n## Backend\n\n### Bluetooth\n\niRobot Education's robots communicate using a common Bluetooth Low Energy serial protocol, the details of which can be found [here](https://github.com/iRobotEducation/root-robot-ble-protocol).\n\n```python\nfrom irobot_edu_sdk.backend.bluetooth import Bluetooth\n\n# Connect to robot over Bluetooth Low Energy.\nbackend0 = Bluetooth('') # Connects to the first BLE robot detected.\nbackend1 = Bluetooth('ROOT') # Use robot named 'ROOT'\n```\n\n## ALPHA features\n**The features detailed below are included as ALPHA versions, which provide limited functionality and are likely to contain several known or unknown bugs.**\n**Support is not provided for ALPHA features at this time.**\n\n### Serial\n\nSome Root® robots support a serial port over USB.\n\n```python\nfrom irobot_edu_sdk.backend.serial import Serial\n\n# Connect to robot over USB-C cable.\nbackend = Serial('COM1')                          # Windows\nbackend = Serial('/dev/cu.usbmodemRT0123F456789') # macOS\nbackend = Serial('/dev/ttyACM0')                  # Linux\n```\n\n### MicroPython\n\nThis SDK requires MicroPython firmware that includes the `uasyncio` library -- version 1.13 or greater.\nDownload the latest release for your board from http://micropython.org/download.\nThen, copy the `irobot_edu_sdk` directory and all contents to the `PYBFLASH` drive.\n\n```python\nfrom irobot_edu_sdk.backend.usb import USB\n\n# Connect to robot from MicroPython-compatible board over USB.\nbackend = USB() # MicroPython-compatible board is powered from USB-C port\n```\n\nNOTE: You may need to use a USB-C to USB-A adapter or a USB-C to micro USB cable.\n\n**© 2022 iRobot Corporation. All rights reserved.**\n* [Terms and Conditions](https://about.irobot.com/en-us/legal/terms-and-conditions)\n* [Privacy Policy](https://edu.irobot.com/privacy-policy)\n\n## Attributions\n* Python® 3 is governed by and a trademark of the Python Software Foundation.\n* The Bluetooth® word mark and logos are registered trademarks owned by Bluetooth SIG, Inc. and any use of such marks by iRobot is under license.\n* Windows® is a trademark of Microsoft Corporation, registered in the U.S. and other countries and regions.\n* macOS® is a trademark of Apple Inc., registered in the U.S. and other countries and regions.\n* Linux® is a trademark of Linus Torvalds, registered in the U.S. and other countries and regions.\n* USB-C™ is a trademark of USB Implementers Forum.\n* All other trademarks mentioned are the property of their respective owners.\n\n\nThis library is made possible thanks to other community software.\nThe following are direct required dependencies:\n* [Bleak](https://github.com/hbldh/bleak), licensed under the MIT license, available [here](https://github.com/hbldh/bleak/blob/develop/LICENSE)\n* [pySerial](https://github.com/pyserial/pyserial/), licensed under the BSD 3-Clause license, available [here](https://github.com/pyserial/pyserial/blob/master/LICENSE.txt)\n* [Root Robot Python Web App](https://github.com/mogenson/root-robot-python-web-app), licensed under the MIT license, available [here](https://github.com/mogenson/root-robot-python-web-app/blob/main/LICENSE)\n\nTo build, the following is recommended:\n* [Poetry](https://github.com/python-poetry/poetry), licensed under the MIT license, available [here](https://github.com/python-poetry/poetry/blob/master/LICENSE)\n\nThe iRobot Education Python SDK provided is licensed under the 3-Clause BSD license, found [here](https://github.com/iRobotEducation/irobot-edu-python-sdk/tree/main/LICENSE.txt)\n",
    'author': 'iRobot Corporation',
    'author_email': 'education@irobot.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/RootRobotics/root-robot-python-sdk',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



#
# Licensed under 3-Clause BSD license available in the License file. Copyright (c) 2020-2022 iRobot Corporation. All rights reserved.
#

from ..packet import Packet


class Backend:
    async def connect(self):
        """Connect to robot"""
        raise NotImplementedError()

    async def is_connected(self) -> bool:
        """Returns True if robot is connected"""
        raise NotImplementedError()

    async def disconnect(self):
        """Disconnect from robot"""
        raise NotImplementedError()

    async def write_packet(self, packet: Packet):
        """Write one packet to the robot"""
        raise NotImplementedError()

    async def read_packet(self) -> Packet:
        """Read one packet from the robot"""
        raise NotImplementedError()



#
# Licensed under 3-Clause BSD license available in the License file. Copyright (c) 2020-2022 iRobot Corporation. All rights reserved.
#

"""
This is a USB class that implements the Backend interface methods.

It is only compatible with a MicroPython board.
"""

from uasyncio import sleep
from binascii import hexlify, unhexlify
from pyb import USB_VCP
from .backend import Backend
from ..packet import Packet


class USB(Backend):
    def __init__(self):
        self._usb = USB_VCP()
        self._usb.init()
        self._usb.setinterrupt(-1)

    async def connect(self):
        await sleep(1)  # pyboard needs a moment to wait for USB
        while not self.is_connected():
            await sleep(0)

    async def is_connected(self) -> bool:
        return self._usb.isconnected()

    async def disconnect(self):
        self._usb.close()

    async def read_packet(self) -> Packet:
        string = b''
        while not (string.endswith(b'\n') and len(string) > 40):
            await sleep(0)
            while not self._usb.any():
                await sleep(0)
            string += self._usb.read(1)
        return Packet.from_bytes(unhexlify(string[-41:-1]))

    async def write_packet(self, packet: Packet):
        string = hexlify(packet.to_bytes()) + b'\n'
        self._usb.write(string)
